"""
Test Abode system setup, shutdown, and general functionality.

Tests the system initialization and attributes of the main Abode system.
"""

import pytest
import requests

import jaraco.abode
import jaraco.abode.devices.status as STATUS
from jaraco.abode import config, settings
from jaraco.abode.helpers import urls

from . import mock as MOCK
from .mock import devices as DEVICES
from .mock import login as LOGIN
from .mock import logout as LOGOUT
from .mock import oauth_claims as OAUTH_CLAIMS
from .mock import panel as PANEL
from .mock import user as USER
from .mock.devices import door_contact as DOOR_CONTACT


@pytest.fixture(autouse=True)
def abode_objects(request):
    self = request.instance
    self.client_no_cred = jaraco.abode.Client()


USERNAME = 'foobar'
PASSWORD = 'deadbeef'


class TestAbode:
    """Test the Abode class."""

    def test_initialization(self):
        """Verify we can initialize abode."""

        assert self.client._username == USERNAME

        assert self.client._password == PASSWORD

    def test_no_credentials(self):
        """Check that we throw an exception when no username/password."""
        with pytest.raises(jaraco.abode.AuthenticationException):
            self.client_no_cred.login()

        self.client_no_cred._username = USERNAME
        with pytest.raises(jaraco.abode.AuthenticationException):
            self.client_no_cred.login()

    def test_manual_login(self, m):
        """Check that we can manually use the login() function."""
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())

        self.client_no_cred.login(username=USERNAME, password=PASSWORD)

    def test_manual_login_with_mfa(self, m):
        """Check that we can login with MFA code."""
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())

        self.client_no_cred.login(username=USERNAME, password=PASSWORD, mfa_code=654321)

    def test_auto_login(self, m):
        """Test that automatic login works."""
        login_json = LOGIN.post_response_ok()
        panel_json = PANEL.get_response_ok()

        m.post(urls.LOGIN, json=login_json)
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.get(urls.PANEL, json=panel_json)
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())

        client = jaraco.abode.Client(
            username='fizz',
            password='buzz',
            auto_login=True,
            get_devices=False,
        )

        assert client._username == 'fizz'
        assert client._password == 'buzz'
        assert client._token == MOCK.AUTH_TOKEN
        assert client._panel == panel_json
        assert client._user == USER.get_response_ok()
        assert client._devices is None
        assert client._automations is None

        client.logout()

        client = None

    def test_auto_fetch(self, m):
        """Test that automatic device and automation retrieval works."""
        login_json = LOGIN.post_response_ok()
        panel_json = PANEL.get_response_ok()

        m.post(urls.LOGIN, json=login_json)
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.get(urls.PANEL, json=panel_json)
        m.get(urls.DEVICES, json=DEVICES.EMPTY_DEVICE_RESPONSE)
        m.get(urls.AUTOMATION, json=DEVICES.EMPTY_DEVICE_RESPONSE)
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())

        client = jaraco.abode.Client(
            username='fizz',
            password='buzz',
            auto_login=False,
            get_devices=True,
            get_automations=True,
        )

        assert client._username == 'fizz'
        assert client._password == 'buzz'
        assert client._token == MOCK.AUTH_TOKEN
        assert client._user == USER.get_response_ok()
        assert client._panel is not None

        # Contains one device, our alarm
        assert client._devices == {'area_1': client.get_alarm()}

        # Contains no automations
        assert client._automations == {}

        client.logout()

        client = None

    def test_login_failure(self, m):
        """Test login failed."""
        m.post(urls.LOGIN, json=LOGIN.post_response_bad_request(), status_code=400)
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())

        # Check that we raise an Exception with a failed login request.
        with pytest.raises(jaraco.abode.AuthenticationException):
            self.client_no_cred.login(username=USERNAME, password=PASSWORD)

    def test_login_mfa_required(self, m):
        """Tests login with MFA code required but not supplied."""
        m.post(
            urls.LOGIN,
            json=LOGIN.post_response_mfa_code_required(),
            status_code=200,
        )

        # Check that we raise an Exception when the MFA code is required
        # but not supplied
        with pytest.raises(jaraco.abode.AuthenticationException):
            self.client_no_cred.login(username=USERNAME, password=PASSWORD)

    def test_login_bad_mfa_code(self, m):
        """Tests login with bad MFA code."""
        m.post(urls.LOGIN, json=LOGIN.post_response_bad_mfa_code(), status_code=400)

        # Check that we raise an Exception with a bad MFA code
        with pytest.raises(jaraco.abode.AuthenticationException):
            self.client_no_cred.login(
                username=USERNAME, password=PASSWORD, mfa_code=123456
            )

    def test_login_unknown_mfa_type(self, m):
        """Tests login with unknown MFA type."""
        m.post(
            urls.LOGIN,
            json=LOGIN.post_response_unknown_mfa_type(),
            status_code=200,
        )

        # Check that we raise an Exception with an unknown MFA type
        with pytest.raises(jaraco.abode.AuthenticationException):
            self.client_no_cred.login(username=USERNAME, password=PASSWORD)

    def test_logout_failure(self, m):
        """Test logout failed."""
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.get(urls.DEVICES, json=DEVICES.EMPTY_DEVICE_RESPONSE)
        m.get(urls.PANEL, json=PANEL.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_bad_request(), status_code=400)

        self.client_no_cred.login(username=USERNAME, password=PASSWORD)

        # Check that we raise an Exception with a failed logout request.
        with pytest.raises(jaraco.abode.AuthenticationException):
            self.client_no_cred.logout()

    def test_logout_exception(self, m):
        """Test logout exception."""
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.get(urls.DEVICES, json=DEVICES.EMPTY_DEVICE_RESPONSE)
        m.get(urls.PANEL, json=PANEL.get_response_ok())
        m.post(urls.LOGOUT, exc=requests.exceptions.ConnectTimeout)

        self.client.login()

        # Check that we eat the exception gracefully
        assert not self.client.logout()

    def test_full_setup(self, m):
        """Check that Abode is set up properly."""
        login_json = LOGIN.post_response_ok()
        panel_json = PANEL.get_response_ok()

        m.post(urls.LOGIN, json=login_json)
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.get(urls.PANEL, json=panel_json)
        m.get(urls.DEVICES, json=DEVICES.EMPTY_DEVICE_RESPONSE)
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())

        self.client.get_devices()

        original_session = self.client._session

        assert self.client._username == USERNAME
        assert self.client._password == PASSWORD
        assert self.client._token == MOCK.AUTH_TOKEN
        assert self.client._user == USER.get_response_ok()
        assert self.client._panel is not None
        assert self.client.get_alarm() is not None
        assert self.client._get_session() is not None
        assert self.client._get_session() == original_session
        assert self.client.events is not None

        self.client.logout()

        assert self.client._token is None
        assert self.client._panel is None
        assert self.client._user is None
        assert self.client._devices is None
        assert self.client._automations is None
        assert self.client._session is not None
        assert self.client._get_session() != original_session

    def test_reauthorize(self, m):
        """Check that Abode can reauthorize after token timeout."""
        new_token = "FOOBAR"
        m.post(
            urls.LOGIN,
            [
                dict(
                    json=LOGIN.post_response_ok(auth_token=new_token),
                    status_code=200,
                ),
            ],
        )

        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())

        m.get(
            urls.DEVICES,
            [
                dict(json=MOCK.response_forbidden(), status_code=403),
                dict(json=DEVICES.EMPTY_DEVICE_RESPONSE, status_code=200),
            ],
        )
        m.get(urls.PANEL, json=PANEL.get_response_ok())

        # Forces a device update
        self.client.get_devices()

        assert self.client._token == new_token

    def test_send_request_exception(self, m):
        """Check that send_request recovers from an exception."""
        new_token = "DEADBEEF"
        m.post(
            urls.LOGIN,
            [
                dict(
                    json=LOGIN.post_response_ok(auth_token=new_token),
                    status_code=200,
                )
            ],
        )

        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())

        m.get(
            urls.DEVICES,
            [
                dict(exc=requests.exceptions.ConnectTimeout),
                dict(json=DEVICES.EMPTY_DEVICE_RESPONSE, status_code=200),
            ],
        )
        m.get(urls.PANEL, json=PANEL.get_response_ok())

        # Forces a device update
        self.client.get_devices()

        assert self.client._token == new_token

    def test_continuous_bad_auth(self, m):
        """Check that Abode won't get stuck with repeated failed retries."""
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.get(urls.DEVICES, json=MOCK.response_forbidden(), status_code=403)

        with pytest.raises(jaraco.abode.Exception):
            self.client.get_devices()

    def test_default_mode(self):
        """Test that the default mode fails if not of type home or away."""
        self.client.set_default_mode('home')
        assert self.client.default_mode == 'home'

        self.client.set_default_mode('away')
        assert self.client.default_mode == 'away'

        with pytest.raises(jaraco.abode.Exception):
            self.client.set_default_mode('foobar')

    def test_all_device_refresh(self, m):
        """Check that device refresh works and reuses the same objects."""
        dc1_devid = 'RF:01'
        dc1a = DOOR_CONTACT.device(devid=dc1_devid, status=STATUS.ON)

        dc2_devid = 'RF:02'
        dc2a = DOOR_CONTACT.device(devid=dc2_devid, status=STATUS.OFF)

        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.DEVICES, json=[dc1a, dc2a])
        m.get(urls.PANEL, json=PANEL.get_response_ok())

        # Reset
        self.client.logout()

        # Get all devices
        self.client.get_devices()

        # Get and check devices

        dc1a_dev = self.client.get_device(dc1_devid)
        assert dc1a['id'] == dc1a_dev.id

        dc2a_dev = self.client.get_device(dc2_devid)
        assert dc2a['id'] == dc2a_dev.id

        # Change device states
        dc1b = DOOR_CONTACT.device(devid=dc1_devid, status=STATUS.OFF)

        dc2b = DOOR_CONTACT.device(devid=dc2_devid, status=STATUS.ON)

        m.get(urls.DEVICES, json=[dc1b, dc2b])

        # Refresh all devices
        self.client.get_devices(refresh=True)

        # Get and check devices again, ensuring they are the same object
        # Future note: "if a is b" tests that the object is the same
        # Thus asserting dc1a_dev is dc1b_dev tests if they are the same object
        dc1b_dev = self.client.get_device(dc1_devid)
        assert dc1b['id'] == dc1b_dev.id
        assert dc1a_dev is dc1b_dev

        dc2b_dev = self.client.get_device(dc2_devid)
        assert dc2b['id'] == dc2b_dev.id
        assert dc2a_dev is dc2b_dev

    def test_settings_validation(self, m):
        """Check that device panel general settings are working."""
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok())
        m.get(urls.SETTINGS, json=MOCK.generic_response_ok())

        with pytest.raises(jaraco.abode.Exception):
            self.client.set_setting("fliptrix", "foobar")

    def test_general_settings(self, m):
        """Check that device panel general settings are working."""
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok())
        m.put(urls.SETTINGS, json=MOCK.generic_response_ok())

        self.client.set_setting(settings.CAMERA_RESOLUTION, settings.CAMERA_RES_640_480)

        self.client.set_setting(settings.CAMERA_GRAYSCALE, settings.ENABLE)

        self.client.set_setting(settings.SILENCE_SOUNDS, settings.ENABLE)

        with pytest.raises(jaraco.abode.Exception):
            self.client.set_setting(settings.CAMERA_RESOLUTION, "foobar")

        with pytest.raises(jaraco.abode.Exception):
            self.client.set_setting(settings.CAMERA_GRAYSCALE, "foobar")

        with pytest.raises(jaraco.abode.Exception):
            self.client.set_setting(settings.SILENCE_SOUNDS, "foobar")

    def test_area_settings(self, m):
        """Check that device panel areas settings are working."""
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok())
        m.put(urls.AREAS, json=MOCK.generic_response_ok())

        self.client.set_setting(
            settings.ENTRY_DELAY_AWAY, settings.ENTRY_EXIT_DELAY_10SEC
        )

        self.client.set_setting(
            settings.EXIT_DELAY_AWAY, settings.ENTRY_EXIT_DELAY_30SEC
        )

        with pytest.raises(jaraco.abode.Exception):
            self.client.set_setting(settings.ENTRY_DELAY_AWAY, "foobar")

        # 10 seconds is invalid here
        with pytest.raises(jaraco.abode.Exception):
            self.client.set_setting(
                settings.EXIT_DELAY_AWAY, settings.ENTRY_EXIT_DELAY_10SEC
            )

    def test_sound_settings(self, m):
        """Check that device panel sound settings are working."""
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok())
        m.put(urls.SOUNDS, json=MOCK.generic_response_ok())

        self.client.set_setting(settings.DOOR_CHIME, settings.SOUND_LOW)

        self.client.set_setting(settings.ALARM_LENGTH, settings.ALARM_LENGTH_2MIN)

        self.client.set_setting(settings.FINAL_BEEPS, settings.FINAL_BEEPS_3SEC)

        with pytest.raises(jaraco.abode.Exception):
            self.client.set_setting(settings.DOOR_CHIME, "foobar")

        with pytest.raises(jaraco.abode.Exception):
            self.client.set_setting(settings.ALARM_LENGTH, "foobar")

        with pytest.raises(jaraco.abode.Exception):
            self.client.set_setting(settings.FINAL_BEEPS, "foobar")

    def test_siren_settings(self, m):
        """Check that device panel siren settings are working."""
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok())
        m.put(urls.SIREN, json=MOCK.generic_response_ok())

        self.client.set_setting(settings.SIREN_ENTRY_EXIT_SOUNDS, settings.ENABLE)

        self.client.set_setting(settings.SIREN_CONFIRM_SOUNDS, settings.ENABLE)

        self.client.set_setting(settings.SIREN_TAMPER_SOUNDS, settings.ENABLE)

        with pytest.raises(jaraco.abode.Exception):
            self.client.set_setting(settings.SIREN_ENTRY_EXIT_SOUNDS, "foobar")

        with pytest.raises(jaraco.abode.Exception):
            self.client.set_setting(settings.SIREN_CONFIRM_SOUNDS, "foobar")

        with pytest.raises(jaraco.abode.Exception):
            self.client.set_setting(settings.SIREN_TAMPER_SOUNDS, "foobar")

    def test_cookies(self, m):
        """Check that cookies are saved and loaded successfully."""
        cookies = dict(SESSION='COOKIE')
        m.post(urls.LOGIN, json=LOGIN.post_response_ok(), cookies=cookies)
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.DEVICES, json=DEVICES.EMPTY_DEVICE_RESPONSE)
        m.get(urls.PANEL, json=PANEL.get_response_ok())

        # Create abode
        client = jaraco.abode.Client(
            username='fizz',
            password='buzz',
            auto_login=False,
            get_devices=False,
        )

        client.login()

        # Test that our cookies are fully realized prior to login

        assert client._session.cookies

        # Test that we now have a cookies file
        cookies_file = config.paths.user_data / 'cookies.json'
        assert cookies_file.exists()

        # Copy the current cookies
        saved_cookies = client._session.cookies

        # New client reads in old data
        client = jaraco.abode.Client(
            username='fizz',
            password='buzz',
            auto_login=False,
            get_devices=False,
        )

        # Test that the cookie data is the same
        assert str(client._session.cookies) == str(saved_cookies)

    def test_empty_cookies(self, m):
        """Check that empty cookies file is loaded successfully."""
        cookies = dict(SESSION='COOKIE')
        m.post(urls.LOGIN, json=LOGIN.post_response_ok(), cookies=cookies)
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.DEVICES, json=DEVICES.EMPTY_DEVICE_RESPONSE)
        m.get(urls.PANEL, json=PANEL.get_response_ok())

        # Create an empty file
        cookie_file = config.paths.user_data / 'cookies.json'

        # Cookies are created
        jaraco.abode.Client(
            username='fizz',
            password='buzz',
            auto_login=True,
            get_devices=False,
        )

        # Test that some cookie data exists
        assert cookie_file.read_bytes()

    def test_invalid_cookies(self, m):
        """Check that empty cookies file is loaded successfully."""
        cookies = dict(SESSION='COOKIE')
        m.post(urls.LOGIN, json=LOGIN.post_response_ok(), cookies=cookies)
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.DEVICES, json=DEVICES.EMPTY_DEVICE_RESPONSE)
        m.get(urls.PANEL, json=PANEL.get_response_ok())

        # Create an invalid pickle file
        config.paths.user_data.joinpath('cookies.json').write_text(
            'invalid cookies', encoding='utf-8'
        )

        # Cookies are created
        empty_client = jaraco.abode.Client(
            username='fizz',
            password='buzz',
            auto_login=True,
            get_devices=False,
        )

        # Test that some cache exists
        assert empty_client._session.cookies



"""Mock Abode User Response."""


def get_response_ok():
    """Return the user response data."""
    return dict(
        id='user@email.com',
        email='user@email.com',
        first_name='John',
        last_name='Doe',
        phone='5555551212',
        profile_pic='https://website.com/default-image.svg',
        address='555 None St.',
        city='New York City',
        state='NY',
        zip='10108',
        country='US',
        longitude='0',
        latitude='0',
        timezone='America/New_York_City',
        verified='1',
        plan='Basic',
        plan_id='0',
        plan_active='1',
        cms_code='1111',
        cms_active='0',
        cms_started_at='',
        cms_expiry='',
        cms_ondemand='',
        step='-1',
        cms_permit_no='',
        opted_plan_id='',
        stripe_account='1',
        plan_effective_from='',
        agreement='1',
        associate_users='1',
        owner='1',
    )



"""Command-line interface."""

import argparse
import code
import contextlib
import getpass
import importlib
import json
import logging
import os
import time

import keyring
from more_itertools import always_iterable

import jaraco.abode
from jaraco.context import suppress
from jaraco.functools import pass_none

from . import Client
from .helpers import timeline as TIMELINE
from .helpers import urls

log = logging.getLogger(__name__)


@suppress(ImportError)
def enable_color():
    import colorlog

    fmt = "%(asctime)s %(levelname)s (%(threadName)s) [%(name)s] %(message)s"
    colorfmt = f"%(log_color)s{fmt}%(reset)s"
    datefmt = '%Y-%m-%d %H:%M:%S'

    handler = logging.getLogger().handlers[0]
    handler.setFormatter(
        colorlog.ColoredFormatter(
            colorfmt,
            datefmt=datefmt,
            reset=True,
            log_colors=dict(
                DEBUG='cyan',
                INFO='green',
                WARNING='yellow',
                ERROR='red',
                CRITICAL='red',
            ),
        )
    )


def setup_logging(log_level=logging.INFO):
    """Set up the logging."""
    logging.basicConfig(level=log_level)

    # Suppress overly verbose logs from libraries that aren't helpful
    logging.getLogger('requests').setLevel(logging.WARNING)
    logging.getLogger('urllib3').setLevel(logging.WARNING)
    logging.getLogger('aiohttp.access').setLevel(logging.WARNING)

    enable_color()
    logging.getLogger().setLevel(log_level)


def build_parser():
    """
    Get parsed arguments.

    >>> parser = build_parser()
    """
    parser = argparse.ArgumentParser("abode")

    parser.add_argument(
        '-u', '--username', help='Username', default=os.environ.get('ABODE_USERNAME')
    )

    parser.add_argument('-p', '--password', help='Password')

    parser.add_argument('--mfa', help='Multifactor authentication code')

    parser.add_argument(
        '--mode',
        help='Output current alarm mode',
        default=False,
        action="store_true",
    )

    parser.add_argument(
        '--arm',
        metavar='mode',
        help='Arm alarm to mode',
    )

    parser.add_argument(
        '--set',
        metavar='setting=value',
        help='Set setting to a value',
        action='append',
    )

    parser.add_argument(
        '--devices',
        help='Output all devices',
        default=False,
        action="store_true",
    )

    parser.add_argument(
        '--device',
        metavar='device_id',
        help='Output one device for device_id',
        action='append',
    )

    parser.add_argument(
        '--interact',
        default=False,
        action="store_true",
        help="interact with the client after running other operations",
    )

    parser.add_argument(
        '--json',
        metavar='device_id',
        help='Output the json for device_id',
        action='append',
    )

    parser.add_argument(
        '--on',
        metavar='device_id',
        help='Switch on a given device_id',
        action='append',
    )

    parser.add_argument(
        '--off',
        metavar='device_id',
        help='Switch off a given device_id',
        action='append',
    )

    parser.add_argument(
        '--lock',
        metavar='device_id',
        help='Lock a given device_id',
        action='append',
    )

    parser.add_argument(
        '--unlock',
        metavar='device_id',
        help='Unlock a given device_id',
        action='append',
    )

    parser.add_argument(
        '--automations',
        help='Output all automations',
        default=False,
        action="store_true",
    )

    parser.add_argument(
        '--activate',
        metavar='automation_id',
        help='Activate (enable) an automation by automation_id',
        action='append',
    )

    parser.add_argument(
        '--deactivate',
        metavar='automation_id',
        help='Deactivate (disable) an automation by automation_id',
        action='append',
    )

    parser.add_argument(
        '--trigger',
        metavar='automation_id',
        help='Trigger (apply) a manual (quick) automation by automation_id',
        action='append',
    )

    parser.add_argument(
        '--capture',
        metavar='device_id',
        help='Trigger a new image capture for the given device_id',
        action='append',
    )

    parser.add_argument(
        '--stream',
        metavar='device_id=location/details.json',
        help='Start a new KVS stream for the given device_id and save the details to a file',
        action='append',
    )

    parser.add_argument(
        '--image',
        metavar='device_id=location/image.jpg',
        help='Save an image from a camera (if available) to the given path',
        action='append',
    )

    parser.add_argument(
        '--listen',
        help='Block and listen for device_id',
        default=False,
        action="store_true",
    )

    parser.add_argument(
        '--debug',
        help='Enable debug logging',
        default=False,
        action="store_true",
    )

    parser.add_argument(
        '--quiet',
        help='Output only warnings and errors',
        default=False,
        action="store_true",
    )

    return parser


def _create_client_instance(args):
    return Client(
        username=args.username,
        password=args.password,
        get_devices=args.mfa is None,
    )


@contextlib.contextmanager
def _log_errors_and_logout(client):
    try:
        yield client
    except jaraco.abode.Exception as exc:
        log.error(exc)
    finally:
        client.logout()


def _device_print(dev, append=''):
    log.info("%s%s", dev.desc, append)


def _device_callback(dev):
    _device_print(dev, ", At: " + time.strftime("%Y-%m-%d %H:%M:%S"))


def _timeline_callback(resp):
    event_code = int(resp['event_code'])
    if 5100 <= event_code <= 5199:
        # Ignore device changes
        return

    tmpl = '{event_name} - {event_type} at {date} {time}'
    log.info(tmpl.format_map(resp))


class Dispatcher:  # pragma: no cover
    def __init__(self, client, args):
        self.client = client
        self.args = args

    def dispatch(self):
        self.login()
        self.output_current_mode()
        self.change_system_mode()
        self.set_setting()
        self.switch_on()
        self.switch_off()
        self.lock()
        self.unlock()
        self.output_json()
        self.print_all_automations()
        self.enable_automation()
        self.disable_automation()
        self.trigger_automation()
        self.trigger_image_capture()
        self.start_kvs_stream()
        self.save_camera_image()
        self.print_all_devices()
        self.print_specific_devices()
        self.start_device_change_listener()
        self.interact()

    def login(self):
        if not self.args.mfa:
            return
        self.client.login(mfa_code=self.args.mfa)
        # fetch devices from Abode
        self.client.get_devices()

    def interact(self):
        if not self.args.interact:
            return

        with contextlib.suppress(ImportError):
            # Allow Up/Down/History in the console
            importlib.import_module('readline')

        banner = 'Interact with the Abode `client`.'
        code.InteractiveConsole(dict(client=self.client)).interact(banner=banner)

    def output_current_mode(self):
        if not self.args.mode:
            return
        log.info("Current alarm mode: %s", self.client.get_alarm().mode)

    def change_system_mode(self):
        if not self.args.arm:
            return
        if self.client.get_alarm().set_mode(self.args.arm):
            log.info("Alarm mode changed to: %s", self.args.arm)
        else:
            log.warning("Failed to change alarm mode to: %s", self.args.arm)

    def set_setting(self):
        for setting in always_iterable(self.args.set):
            key, _, val = setting.partition("=")
            if self.client.set_setting(key, val):
                log.info("Setting %s changed to %s", key, val)

    def _get_device(self, id):
        device = self.client.get_device(id)
        if not device:
            log.warning("Could not find device with id: %s", id)
        return device

    def switch_on(self):
        for device_id in always_iterable(self.args.on):
            self._switch_on(self._get_device(device_id))

    @staticmethod
    @pass_none
    def _switch_on(device):
        if device.switch_on():
            log.info("Switched on device with id: %s", device.id)

    def switch_off(self):
        for device_id in always_iterable(self.args.off):
            self._switch_off(self._get_device(device_id))

    @staticmethod
    @pass_none
    def _switch_off(device):
        if device.switch_off():
            log.info("Switched off device with id: %s", device.id)

    def lock(self):
        for device_id in always_iterable(self.args.lock):
            self._lock(self._get_device(device_id))

    @staticmethod
    @pass_none
    def _lock(device):
        if device.lock():
            log.info("Locked device with id: %s", device.id)

    def unlock(self):
        for device_id in always_iterable(self.args.unlock):
            self._unlock(self._get_device(device_id))

    @staticmethod
    @pass_none
    def _unlock(device):
        if device.unlock():
            log.info("Unlocked device with id: %s", device.id)

    def output_json(self):
        for device_id in always_iterable(self.args.json):
            self._output_json(self._get_device(device_id))

    @staticmethod
    @pass_none
    def _output_json(device):
        print(
            json.dumps(
                device._state,
                sort_keys=True,
                indent=4,
                separators=(',', ': '),
            )
        )

    def print_all_automations(self):
        if not self.args.automations:
            return
        for automation in self.client.get_automations():
            _device_print(automation)

    def _get_automation(self, id):
        automation = self.client.get_automation(id)
        if not automation:
            log.warning("Could not find automation with id: %s", id)
        return automation

    def enable_automation(self):
        for automation_id in always_iterable(self.args.activate):
            self._enable_automation(self._get_automation(automation_id))

    @staticmethod
    @pass_none
    def _enable_automation(automation):
        if automation.enable(True):
            log.info("Activated automation with id: %s", automation.id)

    def disable_automation(self):
        for automation_id in always_iterable(self.args.deactivate):
            self._disable_automation(self._get_automation(automation_id))

    @staticmethod
    @pass_none
    def _disable_automation(automation):
        if automation.enable(False):
            log.info("Deactivated automation with id: %s", automation.id)

    def trigger_automation(self):
        for automation_id in always_iterable(self.args.trigger):
            self._trigger_automation(self._get_automation(automation_id))

    @staticmethod
    @pass_none
    def _trigger_automation(automation):
        if automation.trigger():
            log.info("Triggered automation with id: %s", automation.id)

    def trigger_image_capture(self):
        for device_id in always_iterable(self.args.capture):
            self._trigger_image_capture(self._get_device(device_id))

    @staticmethod
    @pass_none
    def _trigger_image_capture(device):
        if device.capture():
            log.info("Image requested from device with id: %s", device.id)
        else:
            log.warning("Failed to request image from device with id: %s", device.id)

    def start_kvs_stream(self):
        for keyval in always_iterable(self.args.stream):
            dev, _, loc = keyval.partition("=")
            self._start_kvs_stream(self._get_device(dev), loc)

    @staticmethod
    @pass_none
    def _start_kvs_stream(device, path):
        if not device.start_kvs_stream(path):
            log.warning("Failed to start KVS stream for device with id: %s", device.id)

    def save_camera_image(self):
        for keyval in always_iterable(self.args.image):
            dev, _, loc = keyval.partition("=")
            self._save_camera_image(self._get_device(dev), loc)

    @staticmethod
    @pass_none
    def _save_camera_image(device, path):
        try:
            if device.refresh_image() and device.image_to_file(path):
                log.info("Saved image to %s for device id: %s", path, device.id)
        except jaraco.abode.Exception as exc:
            log.warning("Unable to save image: %s", exc)

    def print_all_devices(self):
        if not self.args.devices:
            return
        for device in self.client.get_devices():
            _device_print(device)

    def print_specific_devices(self):
        for device_id in always_iterable(self.args.device):
            device = self._get_device(device_id)
            device and self._print_specific_device(device)

    def _print_specific_device(self, device):
        _device_print(device)

        # Register the specific devices if we decide to listen.
        self.client.events.add_device_callback(device.id, _device_callback)

    def start_device_change_listener(self):
        if not self.args.listen:
            return
        # If no devices were specified then listen to all devices.
        if self.args.device is None:
            log.info("Adding all devices to listener...")

            for device in self.client.get_devices():
                self.client.events.add_device_callback(device.id, _device_callback)

        self.client.events.add_timeline_callback(TIMELINE.ALL, _timeline_callback)

        log.info("Listening for device and timeline updates...")
        self.client.events.start()
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            self.client.events.stop()
            log.info("Device update listening stopped.")


def _get_password(args):
    if not args.username:
        raise SystemExit("Username unknown. Pass a username or set ABODE_USERNAME.")
    args.password = args.password or _get_or_set_password(args.username)


def _get_or_set_password(username):
    password = keyring.get_password(urls.BASE, username)
    if not password:
        password = getpass.getpass(f"Password for {username}: ")
        keyring.set_password(urls.BASE, username, password)
    return password


def main():
    """Execute command line helper."""
    args = build_parser().parse_args()
    _get_password(args)

    setup_logging(log_level=logging.INFO + 10 * (args.quiet - args.debug))

    with _log_errors_and_logout(_create_client_instance(args)) as client:
        Dispatcher(client, args).dispatch()



"""Mock Abode LM (Light/Temp/Humidity) Device."""

DEVICE_ID = 'ZB:eee888801'

TEMP_F = '72 °F'

LUX = '0 lx'

HUMIDITY = '42 %'


def device(
    devid=DEVICE_ID,
    status=TEMP_F,
    temp=TEMP_F,
    lux=LUX,
    humidity=HUMIDITY,
    low_battery=False,
    no_response=False,
):
    """LM mock device."""
    return dict(
        id=devid,
        type_tag="device_type.lm",
        type="LM",
        name="Bedroom Temp",
        area="1",
        zone="15",
        sort_order=None,
        is_window="",
        bypass="0",
        schar_24hr="0",
        sresp_mode_0="0",
        sresp_entry_0="0",
        sresp_exit_0="0",
        sresp_mode_1="5",
        sresp_entry_1="4",
        sresp_exit_1="0",
        sresp_mode_2="0",
        sresp_entry_2="4",
        sresp_exit_2="0",
        sresp_mode_3="0",
        sresp_entry_3="0",
        sresp_exit_3="0",
        version="LMHT_00.00.03.03TC",
        origin="abode",
        deep_link=None,
        status_color="#5cb85c",
        faults=dict(
            low_battery=int(low_battery),
            tempered=0,
            supervision=0,
            out_of_order=0,
            no_response=int(no_response),
        ),
        status=status,
        statusEx="0",
        statuses={
            "hvac_mode": None,
            "temperature": temp,
            "lux": lux,
            "humidity": humidity,
        },
        status_ex="",
        actions=[],
        status_icons=[],
        icon="assets/icons/unknown.svg",
    )



class Numeric(str):
    """
    >>> val = Numeric('item', 3)
    >>> val
    'item'
    >>> int(val)
    3
    """

    def __new__(cls, text, val):
        return super().__new__(cls, text)

    def __init__(self, text, val):
        assert isinstance(val, int)
        self.val = val

    def __int__(self):
        return self.val


ONLINE = 'Online'
OFFLINE = 'Offline'

OPEN = Numeric('Open', 1)
CLOSED = Numeric('Closed', 0)


class Lock:
    OPEN = Numeric('LockOpen', 0)
    CLOSED = Numeric('LockClosed', 1)


ON = Numeric('On', 1)
OFF = Numeric('Off', 0)



"""Test the Abode device classes."""

import pytest

import jaraco.abode
import jaraco.abode.devices.status as STATUS
from jaraco.abode.devices.base import Device
from jaraco.abode.helpers import urls

from .mock import devices as DEVICES
from .mock import login as LOGIN
from .mock import logout as LOGOUT
from .mock import oauth_claims as OAUTH_CLAIMS
from .mock import panel as PANEL
from .mock.devices import door_contact as DOOR_CONTACT
from .mock.devices import glass as GLASS
from .mock.devices import power_switch_sensor as POWERSENSOR
from .mock.devices import unknown as UNKNOWN


class TestDevice:
    """Test the generic device class."""

    def test_device_mapping_type_tag(self):
        """Check new device without type_tag raises exception."""
        device = GLASS.device(
            status=STATUS.ONLINE,
            low_battery=True,
            no_response=True,
            tampered=True,
            out_of_order=True,
        )

        with pytest.raises(jaraco.abode.Exception):
            del device['type_tag']
            Device.new(device, self.client)

    def test_device_auto_naming(self):
        """Check the generic Abode device creates a name."""
        source = GLASS.device(
            status=STATUS.ONLINE,
            low_battery=True,
            no_response=True,
            tampered=True,
            out_of_order=True,
        )

        source['name'] = ""
        device = Device.new(source, self.client)
        generated_name = device.type + ' ' + device.id
        assert device.name == generated_name

        source['name'] = None
        device = Device.new(source, self.client)
        generated_name = device.type + ' ' + device.id
        assert device.name == generated_name

        del source['name']
        device = Device.new(source, self.client)
        generated_name = device.type + ' ' + device.id
        assert device.name == generated_name

    def test_device_init(self, m):
        """Check the generic Abode device class init's properly."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok())

        # Set up device
        source = [
            GLASS.device(
                status=STATUS.ONLINE,
                low_battery=True,
                no_response=True,
                tampered=True,
                out_of_order=True,
                uuid='testuuid00000001',
            )
        ]

        m.get(urls.DEVICES, json=source)

        # Logout to reset everything
        self.client.logout()

        # Get our specific device
        device = self.client.get_device(GLASS.DEVICE_ID)

        # Check device states match
        assert device is not None

        assert device.name == source[0]['name']
        assert device.type == source[0]['type']
        assert device.type_tag == source[0]['type_tag']
        assert device.id == source[0]['id']
        assert device.uuid == source[0]['uuid']
        assert device.status == STATUS.ONLINE
        assert device.battery_low
        assert device.no_response
        assert device.tampered
        assert device.out_of_order
        assert device.desc is not None

    def test_generic_device_refresh(self, m):
        """Check the generic Abode device class init's properly."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok())

        # Set up online device
        m.get(urls.DEVICES, json=[GLASS.device(status=STATUS.ONLINE)])

        # Set up offline device
        device_url = urls.DEVICE.format(id=GLASS.DEVICE_ID)
        m.get(device_url, json=[GLASS.device(status=STATUS.OFFLINE)])

        # Logout to reset everything
        self.client.logout()

        # Get the first device and test
        device = self.client.get_device(GLASS.DEVICE_ID)
        assert device.status == STATUS.ONLINE

        # Refresh the device and test
        device = self.client.get_device(GLASS.DEVICE_ID, refresh=True)
        assert device.status == STATUS.OFFLINE

    def test_multiple_devices(self, m):
        """Tests that multiple devices are returned properly."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))

        # Set up a list of devices
        dev_list = [
            POWERSENSOR.device(),
            DOOR_CONTACT.device(),
            GLASS.device(),
        ]

        m.get(urls.DEVICES, json=dev_list)

        # Logout to reset everything
        self.client.logout()

        # Get our devices
        devices = self.client.get_devices()

        # Assert four devices - three from above + 1 alarm
        assert devices is not None
        assert len(devices) == 4

        # Get each individual device by device ID
        psd = self.client.get_device(POWERSENSOR.DEVICE_ID)
        assert psd is not None

        # Get each individual device by device ID
        psd = self.client.get_device(DOOR_CONTACT.DEVICE_ID)
        assert psd is not None

        # Get each individual device by device ID
        psd = self.client.get_device(GLASS.DEVICE_ID)
        assert psd is not None

    def test_unknown_devices(self, m):
        """Tests that multiple devices are returned properly."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))

        m.get(urls.DEVICES, json=[UNKNOWN.device()])

        # Logout to reset everything
        self.client.logout()

        # Get our devices
        devices = self.client.get_devices()

        # Assert 1 device - skipped device above + 1 alarm
        assert devices is not None
        assert len(devices) == 1

    def test_device_category_filter(self, m):
        """Tests that device category filter returns requested results."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))

        # Set up a list of devices
        dev_list = [
            POWERSENSOR.device(
                devid='ps1',
            ),
            POWERSENSOR.device(
                devid='ps2',
            ),
            GLASS.device(
                devid='gb1',
            ),
        ]

        m.get(urls.DEVICES, json=dev_list)

        # Logout to reset everything
        self.client.logout()

        # Get our glass devices
        devices = self.client.get_devices(generic_type='connectivity')

        assert devices is not None
        assert len(devices) == 1

        # Get our power switch devices
        devices = self.client.get_devices(generic_type='switch')

        assert devices is not None
        assert len(devices) == 2

    def test_no_control_url(self, m):
        """Check that devices return false without control url's."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok())

        m.get(urls.DEVICES, json=GLASS.device(status=STATUS.ONLINE))

        # Logout to reset everything
        self.client.logout()

        # Get device
        device = self.client.get_device(GLASS.DEVICE_ID)

        assert device is not None
        with pytest.raises(jaraco.abode.Exception):
            device.set_status(1)
        with pytest.raises(jaraco.abode.Exception):
            device.set_level('99')

    def test_device_status_changes(self, m):
        """Tests that device status changes work as expected."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(
            urls.DEVICES,
            json=POWERSENSOR.device(),
        )

        # Logout to reset everything
        self.client.logout()

        # Get our power switch
        device = self.client.get_device(POWERSENSOR.DEVICE_ID)

        # Test that we have our device
        assert device is not None
        assert device.status == STATUS.OFF
        assert not device.is_on

        # Set up control url response
        control_url = POWERSENSOR.CONTROL_URL
        m.put(
            control_url,
            json=DEVICES.status_put_response_ok(
                devid=POWERSENSOR.DEVICE_ID, status=int(STATUS.ON)
            ),
        )

        # Change the mode to "on"
        device.switch_on()
        assert device.status == STATUS.ON
        assert device.is_on

        # Change response
        m.put(
            control_url,
            json=DEVICES.status_put_response_ok(
                devid=POWERSENSOR.DEVICE_ID, status=int(STATUS.OFF)
            ),
        )

        # Change the mode to "off"
        device.switch_off()
        assert device.status == STATUS.OFF
        assert not device.is_on

        # Test that an invalid device ID in response throws exception
        m.put(
            control_url,
            json=DEVICES.status_put_response_ok(
                devid='ZW:deadbeef', status=int(STATUS.OFF)
            ),
        )

        with pytest.raises(jaraco.abode.Exception):
            device.switch_on()

        # Test that an invalid status in response throws exception
        m.put(
            control_url,
            json=DEVICES.status_put_response_ok(
                devid=POWERSENSOR.DEVICE_ID, status=int(STATUS.OFF)
            ),
        )

        with pytest.raises(jaraco.abode.Exception):
            device.switch_on()

    def test_device_level_changes(self, m):
        """Tests that device level changes work as expected."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))

        # TODO: Test with a device that supports levels
        m.get(
            urls.DEVICES,
            json=POWERSENSOR.device(),
        )

        # Logout to reset everything
        self.client.logout()

        # Get our power switch
        device = self.client.get_device(POWERSENSOR.DEVICE_ID)

        # Test that we have our device
        assert device is not None
        assert device.status == STATUS.OFF
        assert not device.is_on

        # Set up control url response
        m.put(
            POWERSENSOR.CONTROL_URL,
            json=DEVICES.level_put_response_ok(
                devid=POWERSENSOR.DEVICE_ID, level='100'
            ),
        )

        # Change the level to int 100
        device.set_level(100)
        # self.assertEqual(device.level, '100')

        # Change response
        m.put(
            POWERSENSOR.CONTROL_URL,
            json=DEVICES.level_put_response_ok(devid=POWERSENSOR.DEVICE_ID, level='25'),
        )

        # Change the level to str '25'
        device.set_level('25')
        # self.assertEqual(device.level, '25')

        # Test that an invalid device ID in response throws exception
        m.put(
            POWERSENSOR.CONTROL_URL,
            json=DEVICES.level_put_response_ok(devid='ZW:deadbeef', level='25'),
        )

        with pytest.raises(jaraco.abode.Exception):
            device.set_level(25)

        # Test that an invalid level in response throws exception
        m.put(
            POWERSENSOR.CONTROL_URL,
            json=DEVICES.level_put_response_ok(devid=POWERSENSOR.DEVICE_ID, level='98'),
        )

        with pytest.raises(jaraco.abode.Exception):
            device.set_level('28')

    def test_all_devices(self, all_devices):
        """Tests that all mocked devices are mapped correctly."""

        # Logout to reset everything
        self.client.logout()

        # Loop through all devices
        for device in self.client.get_devices():
            class_type = Device.by_type().get(device.generic_type)

            assert class_type is not None, device.type + ' is not mapped.'
            assert isinstance(device, class_type)

    def test_get_devices_generic_type_substring(self, all_devices):
        """
        Test that a generic_type substring does not match in get_devices.
        """
        assert not self.client.get_devices(generic_type='nect')

    def test_get_devices_generic_type_superstring(self, all_devices):
        """
        Test that a generic_type superstring does not match in get_devices.
        """
        assert not self.client.get_devices(generic_type='connectivity-king')

    def test_get_devices_generic_type_simple_string(self, all_devices):
        """
        Test that a generic_type selects on a simple string.
        """
        selected = self.client.get_devices(generic_type='connectivity')
        all = self.client.get_devices()
        assert selected
        assert len(selected) < len(all)

    def test_get_devices_generic_type_list(self, all_devices):
        """
        Test that a generic_type can select a list of types.
        """
        types = 'door', 'connectivity'
        selected = self.client.get_devices(generic_type=types)
        assert selected
        door_devs = self.client.get_devices(generic_type='door')
        cnct_devs = self.client.get_devices(generic_type='connectivity')
        assert set(selected) == set(door_devs) | set(cnct_devs)



"""Mock Abode Claims Response."""

from . import OAUTH_TOKEN


def get_response_ok(oauth_token=OAUTH_TOKEN):
    """Return the oauth2 claims token."""
    return dict(token_type='Bearer', access_token=oauth_token, expires_in=3600)



"""Mock Abode IR Camera Device."""

import jaraco.abode.devices.status as STATUS

DEVICE_ID = 'ZB:00000005'
CONTROL_URL = 'api/v1/cams/' + DEVICE_ID + '/capture'


def device(devid=DEVICE_ID, status=STATUS.ONLINE, low_battery=False, no_response=False):
    """IR camera mock device."""
    return dict(
        id=devid,
        type_tag='device_type.ir_camera',
        type='Motion Camera',
        name='Downstairs Motion Camera',
        area='1',
        zone='3',
        sort_order=None,
        is_window='',
        bypass='0',
        schar_24hr='0',
        sresp_mode_0='0',
        sresp_entry_0='0',
        sresp_exit_0='0',
        sresp_mode_1='5',
        sresp_entry_1='4',
        sresp_exit_1='0',
        sresp_mode_2='0',
        sresp_entry_2='4',
        sresp_exit_2='0',
        sresp_mode_3='0',
        sresp_entry_3='0',
        sresp_exit_3='0',
        uuid='1234567890',
        version='852_00.00.03.05TC',
        origin='abode',
        control_url='api/v1/cams/ZB:00000005/capture',
        deep_link=None,
        status_color='#5cb85c',
        faults={
            'low_battery': int(low_battery),
            'tempered': 0,
            'supervision': 0,
            'out_of_order': 0,
            'no_response': int(no_response),
        },
        status=status,
        statuses={'hvac_mode': None},
        status_ex='',
        actions=[
            {'label': 'Auto Flash', 'value': 'a=1&z=3&req=img;'},
            {'label': 'Never Flash', 'value': 'a=1&z=3&req=img_nf;'},
        ],
        status_icons=[],
        motion_event='1',
        wide_angle='0',
        icon='assets/icons/motioncamera-b.svg',
    )


def get_capture_timeout():
    """Mock timeout response."""
    return dict(
        code=600, message='Image Capture request has timed out', title='', detail=None
    )


FILE_PATH_ID = 'ZB00000005'
FILE_PATH = 'api/storage/' + FILE_PATH_ID + '/2017-08-23/195505UTC/001.jpg'

LOCATION_HEADER = (
    'https://www.google.com/images/branding/googlelogo/'
    + '1x/googlelogo_color_272x92dp.png'
)


def timeline_event(devid=DEVICE_ID, event_code='5001', file_path=FILE_PATH):
    """Camera Timeline Event Mockup."""
    return dict(
        id=devid,
        event_utc='1503518105',
        nest_activity_zones=None,
        nest_has_motion=None,
        nest_has_sound=None,
        nest_has_person=None,
        date='08/23/2017',
        time='12:55 PM',
        is_alarm='0',
        event_cid='',
        event_code=event_code,
        device_id='ZB:00000005',
        device_type_id='27',
        device_type='Motion Camera',
        device_name='Downstairs Motion Camera',
        file_path=file_path,
        deep_link=None,
        app_deep_link=None,
        file_size='30852',
        file_count='1',
        file_is_del='0',
        event_type='Image Capture',
        severity='6',
        pos='l',
        color='#40bbea',
        viewed_by_uid='',
        user_id='1234',
        user_name='Wil',
        mobile_name='',
        parent_tid='',
        icon='assets/email/motion-camera.png',
        app_type='WebApp',
        file_del_at='',
        event_name='Downstairs Motion Camera Image Capture',
        event_by='',
    )



"""Test the Abode camera class."""

import base64
import json
import os
import pathlib
import re

import pytest

import jaraco.abode
import jaraco.abode.devices.status as STATUS
from jaraco.abode.helpers import urls
from jaraco.collections import Projection

from . import mock as MOCK
from .mock import login as LOGIN
from .mock import logout as LOGOUT
from .mock import oauth_claims as OAUTH_CLAIMS
from .mock import panel as PANEL
from .mock.devices import ipcam as IPCAM
from .mock.devices import ir_camera as IRCAMERA

cam_types = {
    'device_type.ipcam': IPCAM,
    'device_type.ir_camera': IRCAMERA,
}


def all_devices():
    return [
        IRCAMERA.device(),
        IPCAM.device(),
    ]


@pytest.fixture(autouse=True)
def setup_URLs(m):
    # Set up mock URLs
    m.post(urls.LOGIN, json=LOGIN.post_response_ok())
    m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
    m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
    m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
    m.get(urls.DEVICES, json=all_devices())


class TestCamera:
    """Test the camera."""

    def camera_devices(self):
        return (
            device
            for device in self.client.get_devices()
            if device.type_tag != 'device_type.alarm'
        )

    def test_camera_properties(self, m):
        """Tests that camera properties work as expected."""
        for device in self.camera_devices():
            # Specify which device module to use based on type_tag
            cam_type = cam_types[device.type_tag]

            # Test our device
            assert device is not None
            assert device.status == STATUS.ONLINE
            assert not device.battery_low
            assert not device.no_response

            # Set up our direct device get url
            device_url = urls.DEVICE.format(id=device.id)

            # Change device properties
            m.get(
                device_url,
                json=cam_type.device(
                    status=STATUS.OFFLINE,
                    low_battery=True,
                    no_response=True,
                ),
            )

            # Refesh device and test changes
            device.refresh()

            assert device.status == STATUS.OFFLINE
            assert device.battery_low
            assert device.no_response

    def test_camera_capture(self, m):
        """Tests that camera devices capture new images."""
        for device in self.camera_devices():
            # Specify which device module to use based on type_tag
            cam_type = cam_types[device.type_tag]

            # Test that we have the camera devices
            assert device is not None
            assert device.status == STATUS.ONLINE

            # Determine URL based on device type
            if device.type_tag == 'device_type.ipcam':
                url = cam_type.CONTROL_URL_SNAPSHOT

            elif device.type_tag == 'device_type.ir_camera':
                url = cam_type.CONTROL_URL

            # Set up capture URL response
            m.put(url, json=MOCK.generic_response_ok())

            # Capture an image
            assert device.capture()

            # Change capture URL responses
            m.put(url, json=cam_type.get_capture_timeout(), status_code=600)

            # Capture an image with a failure
            assert not device.capture()

    def test_camera_capture_no_control_URLs(self, m):
        """Tests that camera devices capture new images."""
        for device in self.camera_devices():
            # Hide any control URLs from the device state
            device._state = Projection(re.compile('(?!control_url).*'), device._state)

            # Test that jaraco.abode.Exception is raised with no control URLs
            with pytest.raises(jaraco.abode.Exception) as exc:
                device.capture()
            assert exc.value.message == "Control URL does not exist in device JSON."

    def test_camera_image_update(self, m):
        """Tests that camera devices update correctly via timeline request."""
        for device in self.camera_devices():
            # Specify which device module to use based on type_tag
            cam_type = cam_types[device.type_tag]

            # Test that we have our device
            assert device is not None
            assert device.status == STATUS.ONLINE

            # Set up timeline response
            url = urls.TIMELINE_IMAGES_ID.format(device_id=device.id)

            m.get(url, json=[cam_type.timeline_event(device.id)])
            # Set up our file path response
            m.head(
                cam_type.FILE_PATH,
                status_code=302,
                headers={"Location": cam_type.LOCATION_HEADER},
            )

            # Refresh the image
            assert device.refresh_image()

            # Verify the image location
            assert device.image_url == cam_type.LOCATION_HEADER

            # Test that a bad file_path response header results in an exception
            m.head(cam_type.FILE_PATH, status_code=302)

            with pytest.raises(jaraco.abode.Exception):
                device.refresh_image()

            # Test that a bad file_path response code results in an exception
            m.head(
                cam_type.FILE_PATH,
                status_code=200,
                headers={"Location": cam_type.LOCATION_HEADER},
            )

            with pytest.raises(jaraco.abode.Exception):
                device.refresh_image()

            # Test that an an empty timeline event throws exception
            url = urls.TIMELINE_IMAGES_ID.format(device_id=device.id)
            m.get(
                url,
                json=[cam_type.timeline_event(device.id, file_path="")],
            )

            with pytest.raises(jaraco.abode.Exception):
                device.refresh_image()

            # Test that an unexpected timeline event throws exception
            url = urls.TIMELINE_IMAGES_ID.format(device_id=device.id)
            m.get(
                url,
                json=[cam_type.timeline_event(device.id, event_code="1234")],
            )

            with pytest.raises(jaraco.abode.Exception):
                device.refresh_image()

    def test_camera_no_image_update(self, m):
        """Tests that camera updates correctly with no timeline events."""
        for device in self.camera_devices():
            # Test that we have our device
            assert device is not None
            assert device.status == STATUS.ONLINE

            # Set up timeline response
            url = urls.TIMELINE_IMAGES_ID.format(device_id=device.id)
            m.get(url, json=[])

            # Refresh the image
            assert not device.refresh_image()
            assert device.image_url is None

    def test_camera_image_write(self, m):
        """Tests that camera images will write to a file."""
        for device in self.camera_devices():
            # Specify which device module to use based on type_tag
            cam_type = cam_types[device.type_tag]

            # Test that we have our device
            assert device is not None
            assert device.status == STATUS.ONLINE

            # Set up timeline response
            url = urls.TIMELINE_IMAGES_ID.format(device_id=device.id)
            m.get(url, json=[cam_type.timeline_event(device.id)])

            # Set up our file path response
            m.head(
                cam_type.FILE_PATH,
                status_code=302,
                headers={"Location": cam_type.LOCATION_HEADER},
            )

            # Set up our image response
            image_response = "this is a beautiful jpeg image"
            m.get(cam_type.LOCATION_HEADER, json=image_response)

            # Refresh the image
            path = "test.jpg"
            assert device.image_to_file(path, get_image=True)

            # Test the file written and cleanup
            image_data = pathlib.Path(path).read_text(encoding='utf-8')
            assert image_response, image_data
            os.remove(path)

            # Test that bad response returns False
            m.get(cam_type.LOCATION_HEADER, status_code=400)
            with pytest.raises(jaraco.abode.Exception):
                device.image_to_file(path, get_image=True)

            # Test that the image fails to update returns False
            m.get(url, json=[])
            assert not device.image_to_file(path, get_image=True)

    def test_camera_snapshot(self, m):
        """Tests that camera devices capture new snapshots."""
        for device in self.camera_devices():
            # Specify which device module to use based on type_tag
            cam_type = cam_types[device.type_tag]

            # Test that we have our device
            assert device is not None
            assert device.status == STATUS.ONLINE

            # Set up snapshot URL response
            snapshot_url = f"{urls.CAMERA_INTEGRATIONS}{device.uuid}/snapshot"
            m.post(snapshot_url, json=dict(base64Image='test'))

            # Retrieve a snapshot
            assert device.snapshot()

            # Failed snapshot retrieval due to timeout response
            m.post(snapshot_url, json=cam_type.get_capture_timeout(), status_code=600)
            assert not device.snapshot()

            # Failed snapshot retrieval due to missing data
            m.post(snapshot_url, json={})
            assert not device.snapshot()

    def test_camera_snapshot_write(self, m):
        """Tests that camera snapshots will write to a file."""
        for device in self.camera_devices():
            # Specify which device module to use based on type_tag
            cam_type = cam_types[device.type_tag]

            # Test that we have our device
            assert device is not None
            assert device.status == STATUS.ONLINE

            # Set up snapshot URL and image response
            snapshot_url = f"{urls.CAMERA_INTEGRATIONS}{device.uuid}/snapshot"
            image_response = b"this is a beautiful jpeg image"
            b64_image = str(base64.b64encode(image_response), "utf-8")
            m.post(snapshot_url, json=dict(base64Image=b64_image))

            # Request the snapshot and write to file
            path = "test.jpg"
            assert device.snapshot_to_file(path, get_snapshot=True)

            # Test the file written and cleanup
            image_data = pathlib.Path(path).read_bytes()
            assert image_response == image_data
            os.remove(path)

            # Test that bad response returns False
            m.post(snapshot_url, json=cam_type.get_capture_timeout(), status_code=600)
            assert not device.snapshot_to_file(path, get_snapshot=True)

    def test_camera_snapshot_data_url(self, m):
        """Tests that camera snapshots can be converted to a data url."""
        for device in self.camera_devices():
            # Specify which device module to use based on type_tag
            cam_type = cam_types[device.type_tag]

            # Test that we have our device
            assert device is not None
            assert device.status == STATUS.ONLINE

            # Set up snapshot URL and image response
            snapshot_url = f"{urls.CAMERA_INTEGRATIONS}{device.uuid}/snapshot"
            image_response = b"this is a beautiful jpeg image"
            b64_image = str(base64.b64encode(image_response), "utf-8")
            m.post(snapshot_url, json=dict(base64Image=b64_image))

            # Request the snapshot as a data url
            data_url = device.snapshot_data_url(get_snapshot=True)

            # Test the data url matches the image response
            header, encoded = data_url.split(",", 1)
            decoded = base64.b64decode(encoded)
            assert header == "data:image/jpeg;base64"
            assert decoded == image_response

            # Test that bad response returns an empty string
            m.post(snapshot_url, json=cam_type.get_capture_timeout(), status_code=600)
            assert device.snapshot_data_url(get_snapshot=True) == ""

    def test_camera_privacy_mode(self, m):
        """Tests camera privacy mode."""

        # Get the IP camera and test we have it
        device = self.client.get_device(IPCAM.DEVICE_ID)
        assert device is not None
        assert device.status == STATUS.ONLINE

        # Set up params URL response for privacy mode on
        m.put(urls.PARAMS + device.id, json=IPCAM.device(privacy=1))

        # Set privacy mode on
        assert device.privacy_mode(True)

        # Set up params URL response for privacy mode off
        m.put(urls.PARAMS + device.id, json=IPCAM.device(privacy=0))

        # Set privacy mode off
        assert device.privacy_mode(False)

        # Test that an invalid privacy response throws exception
        with pytest.raises(jaraco.abode.Exception):
            device.privacy_mode(True)

    def test_kvs_stream(self, m, tmp_path):
        outfile = tmp_path / 'outfile.txt'
        device = next(self.camera_devices())
        response = dict(channelEndpoint="what goes here?", text="stream info")
        m.post(
            f"{urls.CAMERA_INTEGRATIONS}{device.uuid}/kvs/stream",
            json=response,
        )

        device.start_kvs_stream(outfile)
        assert json.loads(outfile.read_text(encoding='utf-8')) == response



"""Test the Abode device classes."""

import jaraco.abode.devices.status as STATUS
from jaraco.abode.helpers import urls

from .mock import devices as DEVICES
from .mock import login as LOGIN
from .mock import logout as LOGOUT
from .mock import oauth_claims as OAUTH_CLAIMS
from .mock import panel as PANEL
from .mock.devices import secure_barrier as COVER


class TestSecureBarrier:
    """Test the secure barrier class."""

    def test_cover_device_properties(self, m):
        """Tests that cover devices properties work as expected."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(urls.DEVICES, json=COVER.device(status=STATUS.CLOSED))

        # Logout to reset everything
        self.client.logout()

        # Get our power switch
        device = self.client.get_device(COVER.DEVICE_ID)

        # Test our device
        assert device is not None
        assert device.status == STATUS.CLOSED
        assert not device.battery_low
        assert not device.no_response
        assert not device.is_on
        assert not device.is_open

        # Set up our direct device get url
        device_url = urls.DEVICE.format(id=COVER.DEVICE_ID)

        # Change device properties
        m.get(
            device_url,
            json=COVER.device(
                low_battery=True,
                no_response=True,
            ),
        )

        # Refesh device and test changes
        device.refresh()

        assert device.status == STATUS.OPEN
        assert device.battery_low
        assert device.no_response
        assert device.is_on
        assert device.is_open

    def test_cover_status_changes(self, m):
        """Tests that cover device changes work as expected."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))
        m.get(
            urls.DEVICES,
            json=COVER.device(status=STATUS.CLOSED),
        )

        # Logout to reset everything
        self.client.logout()

        # Get our power switch
        device = self.client.get_device(COVER.DEVICE_ID)

        # Test that we have our device
        assert device is not None
        assert device.status == STATUS.CLOSED
        assert not device.is_open

        # Set up control url response
        control_url = urls.BASE + COVER.CONTROL_URL
        m.put(
            control_url,
            json=DEVICES.status_put_response_ok(
                devid=COVER.DEVICE_ID, status=int(STATUS.OPEN)
            ),
        )

        # Change the cover to open
        device.open_cover()
        assert device.status == STATUS.OPEN
        assert device.is_open

        # Change response
        m.put(
            control_url,
            json=DEVICES.status_put_response_ok(
                devid=COVER.DEVICE_ID, status=int(STATUS.CLOSED)
            ),
        )

        # Change the mode to "off"
        device.close_cover()
        assert device.status == STATUS.CLOSED
        assert not device.is_open



import logging

from jaraco.collections import DictAdapter, Projection

from ._itertools import single

log = logging.getLogger(__name__)


class Stateful:
    def __init__(self, state, client):
        """Set up Abode device."""
        self._state = state
        self._client = client

    def __getattr__(self, name):
        try:
            return self._state[name]
        except KeyError as exc:
            raise AttributeError(name) from exc

    def update(self, state):
        """Update the local state from a new state.

        Only updates keys already present.
        """
        self._state.update(Projection(self._state, state))

    @property
    def desc(self):
        """Return a short description of self."""
        return self._desc_t.format_map(DictAdapter(self))

    def refresh(self, path=None):
        """Refresh the device state.

        Useful when not using the notification service.
        """
        tmpl = path or self._url_t
        path = tmpl.format(id=self.id)

        response = self._client.send_request(method="get", path=path)
        state = single(response.json())

        log.debug(f"{self.__class__.__name__} Refresh Response: %s", response.text)

        self._validate(state)
        self.update(state)

        return state

    def _validate(self, state):
        pass



"""Mock Abode Automation."""


def get_response_ok(name: str, enabled: bool, id: str):
    """Return automation json."""
    assert isinstance(enabled, bool)
    return dict(
        name=name,
        enabled=enabled,
        id=id,
        version=2,
        subType='',
        actions=[
            dict(
                directive=dict(
                    trait='panel.trains.panelMode',
                    name='panel.directives.arm',
                    state=dict(panelMode='AWAY'),
                )
            )
        ],
        conditions={},
        triggers=dict(
            operator='OR',
            expressions=list(
                dict(
                    mobileDevices=['89381', '658'],
                    property=dict(
                        trait='mobile.traits.location',
                        name='location',
                        location='31675',
                        equalTo='LAST_OUT',
                    ),
                )
            ),
        ),
    )



"""Timeline event constants."""

import csv

from importlib_resources import files

from jaraco.collections import RangeMap
from jaraco.functools import invoke


class Groups:
    """Event groups."""

    ALARM = 'abode_alarm'
    ALARM_END = 'abode_alarm_end'
    PANEL_FAULT = 'abode_panel_fault'
    PANEL_RESTORE = 'abode_panel_restore'
    DISARM = 'abode_disarm'
    ARM = 'abode_arm'
    ARM_FAULT = 'abode_arm_fault'
    TEST = 'abode_test'
    CAPTURE = 'abode_capture'
    DEVICE = 'abode_device'
    AUTOMATION = 'abode_automation'
    AUTOMATION_EDIT = 'abode_automation_edited'
    ALL = list(locals().values())


event_code_mapping = RangeMap({
    1099: RangeMap.undefined_value,
    1199: Groups.ALARM,
    1299: RangeMap.undefined_value,
    1399: Groups.PANEL_FAULT,
    1499: Groups.DISARM,
    1599: RangeMap.undefined_value,
    1699: Groups.TEST,
    3099: RangeMap.undefined_value,
    3199: Groups.ALARM_END,
    3299: RangeMap.undefined_value,
    3399: Groups.PANEL_RESTORE,
    3799: Groups.ARM,
    3999: RangeMap.undefined_value,
    4001: Groups.ALARM_END,
    4003: Groups.ARM_FAULT,
    4999: RangeMap.undefined_value,
    5099: Groups.CAPTURE,
    5199: Groups.DEVICE,
    5299: Groups.AUTOMATION,
    5999: RangeMap.undefined_value,
    6100: Groups.ARM_FAULT,
})
"""
The best inferred code mapping based on observed events.
Adjust as needed.

>>> event_code_mapping[1100] == Groups.ALARM
True
>>> event_code_mapping[1199] == Groups.ALARM
True
>>> event_code_mapping[1400] == Groups.DISARM
True
>>> event_code_mapping.get(1700)
"""


def map_event_code(event_code):
    """Map a specific event_code to an event group."""
    return event_code_mapping.get(int(event_code))


def _read_events():
    with files().joinpath('events.csv').open(encoding='utf-8') as strm:
        yield from csv.DictReader(strm, quoting=csv.QUOTE_NONE, skipinitialspace=True)


@invoke
def _load_events():
    def var_name(event):
        default = (
            event['text']
            .replace(' - ', '_')
            .replace(' ', '_')
            .replace('/', '_')
            .upper()
        )
        return event['var_name'] or default

    all_events = list(_read_events())
    vars = {
        var_name(event): dict(event_code=event['code'], event_type=event['text'])
        for event in _read_events()
    }
    assert len(all_events) == len(vars)
    globals().update(vars)



"""
An Abode alarm Python library.
"""

import functools
import logging
import uuid

from more_itertools import consume
from requests.exceptions import RequestException
from requests_toolbelt import sessions

import jaraco
from jaraco.collections import Everything
from jaraco.functools import retry
from jaraco.itertools import always_iterable
from jaraco.net.http import cookies

from . import config, settings
from .automation import Automation
from .devices import alarm as ALARM
from .devices.base import Device, Unknown
from .event_controller import EventController
from .exceptions import AuthenticationException
from .helpers import errors as ERROR
from .helpers import urls

log = logging.getLogger(__name__)


@retry(
    retries=1,
    cleanup=lambda: config.paths.user_data.joinpath('cookies.json').unlink(),
    trap=Exception,
)
def _cookies():
    return cookies.ShelvedCookieJar.create(config.paths.user_data)


class Client:
    """Client to an Abode system."""

    def __init__(
        self,
        username=None,
        password=None,
        auto_login=False,
        get_devices=False,
        get_automations=False,
    ):
        self._session = None
        self._token = None
        self._panel = None
        self._user = None
        self._username = username
        self._password = password

        self._event_controller = EventController(self)

        self._default_alarm_mode = 'away'

        self._devices = None

        self._automations = None

        self._session = sessions.BaseUrlSession(urls.BASE)
        self._session.cookies = _cookies()

        if auto_login:
            self.login()

        if get_devices:
            self.get_devices()

        if get_automations:
            self.get_automations()

    def login(self, username=None, password=None, mfa_code=None):
        """Explicit Abode login."""

        self._token = None

        username = username or self._username
        password = password or self._password

        if not isinstance(username, str):
            raise AuthenticationException(ERROR.USERNAME)

        if not isinstance(password, str):
            raise AuthenticationException(ERROR.PASSWORD)

        login_data = {
            'id': username,
            'password': password,
            'uuid': self._session.cookies.get('uuid') or str(uuid.uuid1()),
        }

        if mfa_code is not None:
            login_data['mfa_code'] = mfa_code
            login_data['remember_me'] = 1

        response = self._session.post(urls.LOGIN, json=login_data)
        AuthenticationException.raise_for(response)
        response_object = response.json()

        # Check for multi-factor authentication
        if 'mfa_type' in response_object:
            if response_object['mfa_type'] == "google_authenticator":
                raise AuthenticationException(ERROR.MFA_CODE_REQUIRED)

            raise AuthenticationException(ERROR.UNKNOWN_MFA_TYPE)

        oauth_response = self._session.get(urls.OAUTH_TOKEN)
        AuthenticationException.raise_for(oauth_response)
        oauth_response_object = oauth_response.json()

        log.debug("Login URL: %s", urls.LOGIN)
        log.debug("Login Response: %s", response.text)

        self._token = response_object['token']
        self._panel = response_object['panel']
        self._user = response_object['user']
        self._oauth_token = oauth_response_object['access_token']

        log.info("Login successful")

    def logout(self):
        """Explicit Abode logout."""
        if not self._token:
            return

        header_data = {'ABODE-API-KEY': self._token}

        self._session = sessions.BaseUrlSession(urls.BASE)
        self._token = None
        self._panel = None
        self._user = None
        self._devices = None
        self._automations = None

        try:
            response = self._session.post(urls.LOGOUT, headers=header_data)
        except OSError as exc:
            log.warning("Caught exception during logout: %s", exc)
            return

        AuthenticationException.raise_for(response)

        log.debug("Logout URL: %s", urls.LOGOUT)
        log.debug("Logout Response: %s", response.text)

        log.info("Logout successful")

    def refresh(self):
        """Do a full refresh of all devices and automations."""
        self.get_devices(refresh=True)
        self.get_automations(refresh=True)

    def get_devices(self, refresh=False, generic_type=None):
        """Get all devices from Abode."""
        if refresh or self._devices is None:
            self._load_devices()

        spec_types = (
            Everything() if generic_type is None else set(always_iterable(generic_type))
        )

        return [
            device
            for device in self._devices.values()
            if device.generic_type in spec_types
        ]

    def _load_devices(self):
        if self._devices is None:
            self._devices = {}

        log.info("Updating all devices...")
        response = self.send_request("get", urls.DEVICES)
        devices = always_iterable(response.json())

        log.debug("Get Devices URL (get): %s", urls.AUTOMATION)
        log.debug("Get Devices Response: %s", response.text)

        consume(map(self._load_device, devices))

        # We will be treating the Abode panel itself as an armable device.
        panel_response = self.send_request("get", urls.PANEL)
        panel_json = panel_response.json()

        self._panel.update(panel_json)

        log.debug("Get Mode Panel URL (get): %s", urls.AUTOMATION)
        log.debug("Get Mode Panel Response: %s", response.text)

        alarm_device = self._devices.get(ALARM.id(1))

        if alarm_device:
            alarm_device.update(self._panel)
        else:
            alarm_device = ALARM.create_alarm(self._panel, self)
            self._devices[alarm_device.id] = alarm_device

    def _load_device(self, doc):
        self._reuse_device(doc) or self._create_new_device(doc)

    def _reuse_device(self, doc):
        device = self._devices.get(doc['id'])

        if not device:
            return

        device.update(doc)
        return device

    def _create_new_device(self, doc):
        device = Device.new(doc, self)

        if isinstance(device, Unknown):
            log.debug("Skipping unknown device: %s", doc)
            return

        self._devices[device.id] = device

    def get_device(self, device_id, refresh=False):
        """Get a single device."""
        if self._devices is None:
            self.get_devices()
            refresh = False

        device = self._devices.get(device_id)

        if device and refresh:
            device.refresh()

        return device

    def get_automations(self, refresh=False):
        """Get all automations."""
        if refresh or self._automations is None:
            self._update_all()

        return list(self._automations.values())

    def _update_all(self):
        if self._automations is None:
            # Set up the device libraries
            self._automations = {}

        log.info("Updating all automations...")
        resp = self.send_request("get", urls.AUTOMATION)
        log.debug("Get Automations URL (get): %s", urls.AUTOMATION)
        log.debug("Get Automations Response: %s", resp.text)

        for state in always_iterable(resp.json()):
            # Attempt to reuse an existing automation object
            automation = self._automations.get(str(state['id']))

            # No existing automation, create a new one
            if automation:
                automation.update(state)
            else:
                automation = Automation(state, self)
                self._automations[automation.id] = automation

    def get_automation(self, automation_id, refresh=False):
        """Get a single automation."""
        if self._automations is None:
            self.get_automations()
            refresh = False

        automation = self._automations.get(str(automation_id))

        if automation and refresh:
            automation.refresh()

        return automation

    def get_alarm(self, area='1', refresh=False):
        """Shortcut method to get the alarm device."""
        return self.get_device(ALARM.id(area), refresh)

    def set_default_mode(self, default_mode):
        """Set the default mode when alarms are turned 'on'."""
        if default_mode.lower() not in ('away', 'home'):
            raise jaraco.abode.Exception(ERROR.INVALID_DEFAULT_ALARM_MODE)

        self._default_alarm_mode = default_mode.lower()

    def set_setting(self, name, value, area='1'):
        """Set an abode system setting to a given value."""
        setting = settings.Setting.load(name.lower(), value, area)
        return self.send_request(method="put", path=setting.path, data=setting.data)

    def send_request(self, method, path, headers=None, data=None):
        """Send requests to Abode."""
        attempt = functools.partial(self._send_request, method, path, headers, data)
        return jaraco.functools.retry_call(
            attempt,
            retries=1,
            cleanup=self.login,
            trap=(jaraco.abode.Exception),
        )

    def _send_request(self, method, path, headers, data):
        if not self._token:
            self.login()

        if not headers:
            headers = {}

        headers['Authorization'] = 'Bearer ' + self._oauth_token
        headers['ABODE-API-KEY'] = self._token

        try:
            response = getattr(self._session, method)(path, headers=headers, json=data)

            if response and response.status_code < 400:
                return response
        except RequestException:
            log.info("Abode connection reset...")

        raise jaraco.abode.Exception(ERROR.REQUEST)

    @property
    def default_mode(self):
        """Get the default mode."""
        return self._default_alarm_mode

    @property
    def events(self):
        """Get the event controller."""
        return self._event_controller

    @property
    def uuid(self):
        """Get the UUID."""
        return self._session.cookies['uuid']

    def _get_session(self):
        # Perform a generic update so we know we're logged in
        self.send_request("get", urls.PANEL)

        return self._session



"""Test the Abode binary sensors."""

import functools
import itertools

import jaraco.abode.devices.status as STATUS
from jaraco.abode.devices import base
from jaraco.abode.helpers import urls

from .mock import login as LOGIN
from .mock import logout as LOGOUT
from .mock import oauth_claims as OAUTH_CLAIMS
from .mock import panel as PANEL
from .mock.devices import door_contact as DOOR_CONTACT
from .mock.devices import glass as GLASS
from .mock.devices import keypad as KEYPAD
from .mock.devices import occupancy as OCCUPANCY
from .mock.devices import pir as PIR
from .mock.devices import remote_controller as REMOTE_CONTROLLER
from .mock.devices import siren as SIREN
from .mock.devices import status_display as STATUS_DISPLAY
from .mock.devices import water_sensor as WATER_SENSOR


class TestBinarySensors:
    """Test the binary sensors."""

    def test_binary_sensor_properties(self, m):
        """Tests that binary sensor device properties work as expected."""
        # Set up URLs
        m.post(urls.LOGIN, json=LOGIN.post_response_ok())
        m.get(urls.OAUTH_TOKEN, json=OAUTH_CLAIMS.get_response_ok())
        m.post(urls.LOGOUT, json=LOGOUT.post_response_ok())
        m.get(urls.PANEL, json=PANEL.get_response_ok(mode='standby'))

        # Set up all Binary Sensor Devices in "off states"
        all_devices = [
            DOOR_CONTACT.device(),
            GLASS.device(
                status=STATUS.OFFLINE,
            ),
            KEYPAD.device(
                status=STATUS.OFFLINE,
            ),
            OCCUPANCY.device(),
            PIR.device(
                status=STATUS.OFFLINE,
            ),
            REMOTE_CONTROLLER.device(
                status=STATUS.OFFLINE,
            ),
            SIREN.device(
                status=STATUS.OFFLINE,
            ),
            STATUS_DISPLAY.device(
                status=STATUS.OFFLINE,
            ),
            WATER_SENSOR.device(),
        ]

        m.get(urls.DEVICES, json=all_devices)

        # Logout to reset everything
        self.client.logout()

        # Test our devices
        for device in self.client.get_devices():
            assert not device.is_on, device.type + " is_on failed"
            assert not device.battery_low, device.type + " battery_low failed"
            assert not device.no_response, device.type + " no_response failed"

        # Set up all Binary Sensor Devices in "off states"
        all_devices = [
            DOOR_CONTACT.device(
                status=STATUS.OPEN,
                low_battery=True,
                no_response=True,
            ),
            GLASS.device(
                low_battery=True,
                no_response=True,
            ),
            KEYPAD.device(
                low_battery=True,
                no_response=True,
            ),
            OCCUPANCY.device(
                has_motion=True,
                low_battery=True,
                no_response=True,
            ),
            PIR.device(
                low_battery=True,
                no_response=True,
            ),
            REMOTE_CONTROLLER.device(
                low_battery=True,
                no_response=True,
            ),
            SIREN.device(
                low_battery=True,
                no_response=True,
            ),
            STATUS_DISPLAY.device(
                low_battery=True,
                no_response=True,
            ),
            WATER_SENSOR.device(
                status=STATUS.ON,
                low_battery=True,
                no_response=True,
            ),
        ]

        m.get(urls.DEVICES, json=all_devices)

        # Refesh devices and test changes
        for device in skip_alarms(self.client.get_devices(refresh=True)):
            assert device.is_on, device.type + " is_on failed"
            assert device.battery_low, device.type + " battery_low failed"
            assert device.no_response, device.type + " no_response failed"


def test_binary_sensor_classes():
    device = base.Device.new(OCCUPANCY.device(), None)
    assert device.generic_type == 'motion'


def is_alarm(device):
    return device.type_tag == 'device_type.alarm'


skip_alarms = functools.partial(itertools.filterfalse, is_alarm)
